package dev.pfaff.glass.glfw;

import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.lang.foreign.Arena;
import java.lang.foreign.FunctionDescriptor;
import java.lang.foreign.Linker;
import java.lang.foreign.MemorySegment;
import java.lang.invoke.MethodHandle;
import java.nio.charset.StandardCharsets;
import java.util.Objects;

import static dev.pfaff.glass.glfw.GLFW.*;
import static java.lang.foreign.ValueLayout.*;

/**
 * @see <a href="https://www.glfw.org/docs/3.4/group__window.html">Window reference</a>
 */
public record GLFWwindow(GLFW glfw, MemorySegment handle) {
	public GLFWwindow {
		Objects.requireNonNull(glfw, "glfw cannot be null");
		Objects.requireNonNull(handle, "handle cannot be null");
	}

	public static final MethodHandle MH_glfwMakeContextCurrent = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow));
	public static final MethodHandle MH_glfwGetCurrentContext = LINKER.downcallHandle(FunctionDescriptor.of(ML_GLFWwindow),
			Linker.Option.critical(false));
	public static final MethodHandle MH_glfwSwapInterval = LINKER.downcallHandle(FunctionDescriptor.ofVoid(JAVA_INT));
	public static final MethodHandle MH_glfwExtensionSupported = LINKER.downcallHandle(
			FunctionDescriptor.of(JAVA_BOOLEAN, ADDRESS.withTargetLayout(JAVA_BYTE)));
	public static final MethodHandle MH_glfwCreateWindow = LINKER.downcallHandle(
			FunctionDescriptor.of(ML_GLFWwindow,
					JAVA_INT,
					JAVA_INT,
					ADDRESS.withTargetLayout(JAVA_BYTE),
					ML_GLFWmonitor,
					ML_GLFWwindow));
	public static final MethodHandle MH_glfwWindowHint = LINKER.downcallHandle(FunctionDescriptor.ofVoid(JAVA_INT, JAVA_INT));
	public static final MethodHandle MH_glfwDestroyWindow = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow));
	public static final MethodHandle MH_glfwWindowShouldClose = LINKER.downcallHandle(FunctionDescriptor.of(JAVA_BOOLEAN, ML_GLFWwindow),
			Linker.Option.critical(false));
	public static final MethodHandle MH_glfwGetFramebufferSize = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow, ADDRESS.withTargetLayout(JAVA_INT), ADDRESS.withTargetLayout(JAVA_INT)),
			Linker.Option.critical(true));
	public static final MethodHandle MH_glfwGetWindowSize = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow, ADDRESS.withTargetLayout(JAVA_INT), ADDRESS.withTargetLayout(JAVA_INT)),
			Linker.Option.critical(true));
	public static final MethodHandle MH_glfwGetWindowContentScale = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow, ADDRESS.withTargetLayout(JAVA_FLOAT), ADDRESS.withTargetLayout(JAVA_FLOAT)),
			Linker.Option.critical(true));
	public static final MethodHandle MH_glfwGetWindowAttrib = LINKER.downcallHandle(FunctionDescriptor.of(JAVA_INT, ML_GLFWwindow, JAVA_INT),
			Linker.Option.critical(true));
	public static final MethodHandle MH_glfwGetClipboardString = LINKER.downcallHandle(FunctionDescriptor.of(ADDRESS.withTargetLayout(JAVA_BYTE), ML_GLFWwindow));
	public static final MethodHandle MH_glfwSetClipboardString = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow, ADDRESS.withTargetLayout(JAVA_BYTE)));
	public static final MethodHandle MH_glfwSwapBuffers = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow));
	public static final MethodHandle MH_glfwShowWindow = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow));
	public static final MethodHandle MH_glfwHideWindow = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow));
	public static final MethodHandle MH_glfwSetWindowSize = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow, JAVA_INT, JAVA_INT));

	public static void setIntegerHint(GLFW glfw, WindowHint hint, int value) {
		try {
			glfw.MH_glfwWindowHint_bound.invokeExact(hint.value(), value);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public static void setBooleanHint(GLFW glfw, WindowHint hint, boolean value) {
		setIntegerHint(glfw, hint, value ? 1 : 0);
	}

	@NonNull
	public static GLFWwindow create(GLFW glfw, int width, int height, String title, @Nullable GLFWmonitor monitor, @Nullable GLFWwindow share) {
		var monitorHandle = monitor != null ? monitor.handle() : MemorySegment.NULL;
		var shareHandle = share != null ? share.handle : MemorySegment.NULL;
		MemorySegment handle;
		try (var arena = Arena.ofConfined()) {
			var titlePtr = arena.allocateFrom(title, StandardCharsets.UTF_8);
			try {
				handle = (MemorySegment) glfw.MH_glfwCreateWindow_bound.invokeExact(width, height, titlePtr, monitorHandle, shareHandle);
			} catch (Throwable e) {
				throw new RuntimeException(e);
			}
		}
		if (handle.equals(MemorySegment.NULL)) throw new RuntimeException("Unable to create GLFW window");
		return new GLFWwindow(glfw, handle);
	}

	@Nullable
	public static GLFWwindow getCurrentContext(GLFW glfw) {
		MemorySegment handle;
		try {
			handle = (MemorySegment) glfw.MH_glfwGetCurrentContext_bound.invokeExact();
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
		if (handle.equals(MemorySegment.NULL)) return null;
		return new GLFWwindow(glfw, handle);
	}

	public void makeContextCurrent() {
		try {
			glfw.MH_glfwMakeContextCurrent_bound.invokeExact(handle);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void destroy() {
		try {
			glfw.MH_glfwDestroyWindow_bound.invokeExact(handle);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public boolean shouldClose() {
		try {
			return (boolean) glfw.MH_glfwWindowShouldClose_bound.invokeExact(handle);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void setVisibility(boolean value) {
		try {
			if (value) {
				glfw.MH_glfwShowWindow_bound.invokeExact(handle);
			} else {
				glfw.MH_glfwHideWindow_bound.invokeExact(handle);
			}
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public record WindowSize(int width, int height) {}
	public record FramebufferSize(int width, int height) {}
	public record ContentScale(float x, float y) {
		public float min() {
			return Math.min(x, y);
		}

		public float max() {
			return Math.max(x, y);
		}
	}

	public WindowSize getWindowSize() {
		try {
			var slot = new long[1];
			var buf = MemorySegment.ofArray(slot);
			glfw.MH_glfwGetWindowSize_bound.invokeExact(handle, buf, buf.asSlice(JAVA_INT.byteSize()));
			return new WindowSize(buf.getAtIndex(JAVA_INT, 0), buf.getAtIndex(JAVA_INT, 1));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public FramebufferSize getFramebufferSize() {
		try {
			var slot = new long[1];
			var buf = MemorySegment.ofArray(slot);
			glfw.MH_glfwGetFramebufferSize_bound.invokeExact(handle, buf, buf.asSlice(JAVA_INT.byteSize()));
			return new FramebufferSize(buf.getAtIndex(JAVA_INT, 0), buf.getAtIndex(JAVA_INT, 1));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public ContentScale getWindowContentScale() {
		try {
			var slot = new long[1];
			var buf = MemorySegment.ofArray(slot);
			glfw.MH_glfwGetWindowContentScale_bound.invokeExact(handle, buf, buf.asSlice(JAVA_FLOAT.byteSize()));
			return new ContentScale(buf.getAtIndex(JAVA_FLOAT, 0), buf.getAtIndex(JAVA_FLOAT, 1));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public int getWindowAttrib(WindowAttrib attrib) {
		try {
			return (int) glfw.MH_glfwGetWindowAttrib_bound.invokeExact(handle, attrib.value());
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public boolean getBooleanWindowAttrib(WindowAttrib attrib) {
		int value = getWindowAttrib(attrib);
		if (value == 0) return false;
		if (value == 1) return true;
		throw new IllegalStateException("GLFW returned invalid boolean");
	}

	/**
	 * @return the contents of the clipboard if UTF-8 encoded or convertible, or {@code null} if empty or inconvertible.
	 */
	@Nullable
	public String getClipboardString() {
		glfw.requireOnMainThread();
		try {
			var ptr = (MemorySegment) glfw.MH_glfwGetClipboardString_bound.invokeExact(handle);
			return ptr.reinterpret(Long.MAX_VALUE).getString(0, StandardCharsets.UTF_8);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void setClipboardString(String string) {
		glfw.requireOnMainThread();
		try (var arena = Arena.ofConfined()) {
			var ptr = arena.allocateFrom(string, StandardCharsets.UTF_8);
			glfw.MH_glfwSetClipboardString_bound.invokeExact(handle, ptr);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void swapBuffers() {
		try {
			glfw.MH_glfwSwapBuffers_bound.invokeExact(handle);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public static void setSwapInterval(GLFW glfw, int interval) {
		try {
			glfw.MH_glfwSwapInterval_bound.invokeExact(interval);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void setWindowSize(int width, int height) {
		try {
			glfw.MH_glfwSetWindowSize_bound.invokeExact(handle, width, height);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * @return whether the named extension is supported by the current context.
	 */
	public static boolean isExtensionSupported(GLFW glfw, String name) {
		try (var arena = Arena.ofConfined()) {
			var namePtr = arena.allocateFrom(name, StandardCharsets.US_ASCII);
			try {
				return (boolean) glfw.MH_glfwExtensionSupported_bound.invokeExact(namePtr);
			} catch (Throwable e) {
				throw new RuntimeException(e);
			}
		}
	}
}
